//this file is part of eMule
//Copyright (C)2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#pragma once
#include "AsyncSocketEx.h"
#include "OtherFunctions.h"
#include "ThrottledSocket.h" // ZZ:UploadBandWithThrottler (UDP)

#include "NetF/NetF_Tick.h" // netfinity: Type safe timers
//#include "NetF/Messages.h" // netfinity: Inter-thread message queues
#include "NetF/SafePointers.h" // netfinity: Safe deletion of sockets

class CAsyncProxySocketLayer;
class Packet;

#define ERR_WRONGHEADER		0x01
#define ERR_TOOBIG			0x02

#define	ES_DISCONNECTED		0xFF
#define	ES_NOTCONNECTED		0x00
#define	ES_CONNECTED		0x01

#define PACKET_HEADER_SIZE	6

struct StandardPacketQueueEntry {
    uint32 actualPayloadSize;
    Packet* packet;
};

class CEMSocket : public CAsyncSocketEx, public ThrottledFileSocket, public CSafeObject // ZZ:UploadBandWithThrottler (UDP)
{
	DECLARE_DYNAMIC(CEMSocket)

public:
	CEMSocket();
	virtual ~CEMSocket();

	virtual void 	SendPacket(Packet* packet, bool delpacket = true, bool controlpacket = true, uint32 actualPayloadSize = 0);
    bool    HasQueues();
    bool	IsConnected() const {return byConnected == ES_CONNECTED;}
	uint8	GetConState() const {return byConnected;}
	virtual bool IsRawDataMode() const { return false; }
	//void	SetDownloadLimit(const CBytes& limit);
	void	MOD_SetDownloadLimitByRate(const CBytesPerSecond& limit); // netfinty: More accurate rate limiting
	void	MOD_RefreshDownloadLimit();
	CBytes	MOD_GetDownloadLimit();
	void	MOD_SetUsedDownloadLimit(const CBytes& bytes);
	void	MOD_SetDownloadRate(const CBytesPerSecond& datarate);
	void	DisableDownloadLimit();
	CBytes	MOD_GetReceivedBytesSinceLastCallAndReset();
	// MOD BEGIN netfinity: Upload (pseudo) rate control
	void	MOD_RefreshUploadLimit();
	CBytes	MOD_GetUploadLimit();
	void	MOD_SetUsedUploadLimit(const CBytes& bytes);
	void	MOD_SetUploadRate(const CBytesPerSecond& datarate);
	void	MOD_DisableUploadLimit();
	// MOD END netfinity
	BOOL	AsyncSelect(long lEvent);

	virtual CMilliSeconds GetTimeOut() const;
	virtual void SetTimeOut(const CMilliSeconds& uTimeOut);

	virtual BOOL Connect(LPCTSTR lpszHostAddress, UINT nHostPort);
	virtual BOOL Connect(SOCKADDR* pSockAddr, int iSockAddrLen);
	void InitProxySupport();
	virtual void RemoveAllLayers();

    CTick GetLastCalledSend() { return lastCalledSend; }

    CBytes GetSentBytesCompleteFileSinceLastCallAndReset();
    CBytes GetSentBytesPartFileSinceLastCallAndReset();
    CBytes GetSentBytesControlPacketSinceLastCallAndReset();
    CBytes GetSentPayloadSinceLastCallAndReset();
    void TruncateQueues();

    virtual SocketSentBytes SendControlData(CBytes maxNumberOfBytesToSend, CBytes minFragSize) { return Send(maxNumberOfBytesToSend, minFragSize, true); };
    virtual SocketSentBytes SendFileAndControlData(CBytes maxNumberOfBytesToSend, CBytes minFragSize, bool overrideLimit = false) { return Send(maxNumberOfBytesToSend, minFragSize, false, overrideLimit); }; 	// netfinity: Upload (pseudo) rate control

    CBytes	GetNeededBytes();

#ifdef _DEBUG
	// Diagnostic Support
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:
	virtual int	OnLayerCallback(const CAsyncSocketExLayer *pLayer, int nType, int nParam1, int nParam2);
	
	virtual void	DataReceived(const BYTE* pcData, UINT uSize);
public:
	virtual bool	PacketReceived(Packet* packet) = 0;
protected:
	virtual void	OnError(int nErrorCode) = 0;
	virtual void	OnClose(int nErrorCode);
	virtual void	OnSend(int nErrorCode);	
	virtual void	OnReceive(int nErrorCode);

	uint8			byConnected;
	CMilliSeconds	m_uTimeOut;
	bool	m_ProxyConnectFailed;
	CAsyncProxySocketLayer* m_pProxyLayer;

private:
    virtual SocketSentBytes Send(CBytes maxNumberOfBytesToSend, CBytes minFragSize, bool onlyAllowedToSendControlPacket, bool overrideLimit = false); 	// netfinity: Upload (pseudo) rate control
	void	ClearQueues();
	virtual int Receive(void* lpBuf, int nBufLen, int nFlags = 0);

    CBytes GetNextFragSize(CBytes current, CBytes minFragSize);
    bool    HasSent() { return m_hasSent; }

	// Download (pseudo) rate control	
	CBytesPerSecond	downloadRate;
	CBytes			downloadLimit;
	bool			downloadLimitEnable;
	CTick			downloadLimitLastCall; // netfinity: Rate based limiting
	bool			pendingOnReceive;
	// Keep track of bytes received on the socket
	CBytes			m_actualBytesReceived;
	// MOD BEGIN netfinity: Upload (pseudo) rate control
	CBytesPerSecond	uploadRate;
	CBytes			uploadLimit;
	bool			uploadLimitEnable;
	CTick			uploadLimitLastCall;
	// MOD END netfinity

	// Download partial header
	char	pendingHeader[PACKET_HEADER_SIZE];	// actually, this holds only 'PACKET_HEADER_SIZE-1' bytes.
	uint32	pendingHeaderSize;

	// Download partial packet
	Packet* pendingPacket;
	uint32  pendingPacketSize;

	// Receive thread locking
    CCriticalSection receiveLocker;

	// Upload control
	char*	sendbuffer;
	CBytes	sendblen;
	CBytes	sent;

	CTypedPtrList<CPtrList, Packet*> controlpacket_queue;
	CList<StandardPacketQueueEntry> standartpacket_queue;
    bool m_currentPacket_is_controlpacket;
    CCriticalSection sendLocker;
    CBytes m_numberOfSentBytesCompleteFile;
    CBytes m_numberOfSentBytesPartFile;
    CBytes m_numberOfSentBytesControlPacket;
    bool m_currentPackageIsFromPartFile;
	bool m_bAccelerateUpload;
    CTick	lastCalledSend;
    CTick	lastSent;
	CTick	lastFinishedStandard;
    CBytes m_actualPayloadSize;
    CBytes m_actualPayloadSizeSent;
    bool m_bBusy;
    bool m_hasSent;
};
