//this file is part of eMule
//Copyright (C)2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

/** @file	IOThread.cpp
 *  @brief	Disk IO handling thread
 *  @author	netfinity
 */
#pragma once
#include "stdafx.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#include "emule.h"
#include "emuleDlg.h"
#include "Packets.h"
#include "PartFile.h"
#include "log.h"
#include "NetF.h"
#include "SafePointers.h"
#include "Messages.h"

#include "IOThread.h"

//IMPLEMENT_DYNCREATE(CIOThread, CWinThread)

CIOThread disk_io_thread;

CIOThread::CIOThread()
{
	CreateThread();
}

CIOThread::~CIOThread()
{
}

BOOL CIOThread::InitInstance()
{
	DbgSetThreadName("DiskIOThread");
	InitThreadLocale();
	return TRUE;
}

BOOL CIOThread::Run()
{
	CMessage*			pMsg;
	CList<CMessage*>	listQueued;
	bool				bDeleteMsg;
	while (1) // Shouldn't we exit the loop when eMule is closing!? (Too lacy!)
	{
		WaitMessage();
		while (pMsg = GetMessage())
		{
			bDeleteMsg = true;
			switch (pMsg->GetID())
			{
				case CIOReadReq_Msg::ID:
				{
					CIOReadReq_Msg* pCastMsg = static_cast<CIOReadReq_Msg*>(pMsg);
					try
					{
						if (IORead(pCastMsg->GetFile(), pCastMsg->GetReqblock(), pCastMsg->GetBuffer()))
							pCastMsg->Reply(new CIOReadRsp_Msg(pCastMsg->GetClient(), pCastMsg->GetFile(), pCastMsg->GetReqblock(), pCastMsg->GetBuffer()));
						else
						{
							listQueued.AddTail(pMsg);
							bDeleteMsg = false;
						}
					}
					catch(CString error)
					{
						pCastMsg->Reply(new CIOReadErr_Msg(pCastMsg->GetClient(), pCastMsg->GetFile(), pCastMsg->GetReqblock(), pCastMsg->GetBuffer(), error));
					}
					catch(CFileException* e)
					{
						TCHAR szError[MAX_CFEXP_ERRORMSG];
						e->GetErrorMessage(szError, ARRSIZE(szError));
						e->Delete();
						pCastMsg->Reply(new CIOReadErr_Msg(pCastMsg->GetClient(), pCastMsg->GetFile(), pCastMsg->GetReqblock(), pCastMsg->GetBuffer(), szError));
					}
					catch(...)
					{
						pCastMsg->Reply(new CIOReadErr_Msg(pCastMsg->GetClient(), pCastMsg->GetFile(), pCastMsg->GetReqblock(), pCastMsg->GetBuffer(), _T("Major screwup")));
					}
					break;
				}
				default:
				{
					break;
				}
			}
			if (bDeleteMsg)
				delete pMsg;
		}

		while (!listQueued.IsEmpty())
		{
			this->PostMessage(listQueued.RemoveHead());
		}
		Sleep(CMilliSeconds(100));
	}
	return 0;
}

BOOL CIOThread::IORead(CKnownFile* srcfile, Requested_Block_Struct* currentblock, byte* buffer)
{
    CFile file;
	CString fullname;
	bool bFromPF = true; // Statistic to breakdown uploaded data by complete file vs. partfile.
	CSingleLock lockFile(&((CPartFile*)srcfile)->m_FileCompleteMutex);

	uint32 togo;
	if (currentblock->StartOffset > currentblock->EndOffset)
		togo = currentblock->EndOffset + (srcfile->GetFileSize() - currentblock->StartOffset);
	else
		togo = currentblock->EndOffset - currentblock->StartOffset;
	
	if (!srcfile)
		throw GetResString(IDS_ERR_REQ_FNF);

	if (srcfile->IsPartFile() && ((CPartFile*)srcfile)->GetStatus() != PS_COMPLETE){
		// Do not access a part file, if it is currently moved into the incoming directory.
		if (!lockFile.Lock(0)){ // just do a quick test of the mutex's state and return if it's locked.
			return false;
		}
		// If it's a part file which we are uploading the file remains locked until we've read the
		// current block. This way the file completion thread can not (try to) "move" the file into
		// the incoming directory.

		fullname = RemoveFileExtension(((CPartFile*)srcfile)->GetFullName());
	}
	else{
		fullname.Format(_T("%s\\%s"),srcfile->GetPath(),srcfile->GetFileName());
	}
	
	if (!srcfile->IsPartFile()){
		bFromPF = false; // This is not a part file...
		if (!file.Open(fullname,CFile::modeRead|CFile::osSequentialScan|CFile::shareDenyNone))
			throw GetResString(IDS_ERR_OPEN);
		file.Seek(currentblock->StartOffset,0);
			
		if (uint32 done = file.Read(buffer,togo) != togo){
			file.SeekToBegin();
			file.Read(buffer + done,togo-done);
		}
		file.Close();
	}
	else{
		CPartFile* partfile = (CPartFile*)srcfile;

		partfile->m_hpartfile.Seek(currentblock->StartOffset,0);
		
		if (uint32 done = partfile->m_hpartfile.Read(buffer,togo) != togo){
			partfile->m_hpartfile.SeekToBegin();
			partfile->m_hpartfile.Read(buffer + done,togo-done);
		}
	}
	return true;
}


BEGIN_MESSAGE_MAP(CIOThread, CWinThread)
END_MESSAGE_MAP()