//this file is part of eMule
//Copyright (C)2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <stdafx.h>
#include "NetF.h"
#include "Preferences.h"
#include "ini2.h"

CNetF	theNetF;

CNetF::CNetF()
{
	// Status
	m_bHighIdPossible = false;
	// Statistics
	m_iSessionSuccessfulDownloadNNP = 0;
	m_iSessionSuccessfulDownloadLimit = 0;
	m_iSessionSuccessfulDownloadTimeout = 0;
	m_iSessionSuccessfulDownloadDisconnect = 0;
	m_iSessionSuccessfulDownloadDropped = 0;
	m_iSessionSuccessfulDownloadOther = 0;
	m_iSessionFailedDownloadNNP = 0;
	m_iSessionFailedDownloadLimit = 0;
	m_iSessionFailedDownloadTimeout = 0;
	m_iSessionFailedDownloadDisconnect = 0;
	m_iSessionFailedDownloadDropped = 0;
	m_iSessionFailedDownloadOther = 0;
	m_iSessionSuccessfulUploadCancel = 0;
	m_iSessionSuccessfulUploadTimeout = 0;
	m_iSessionSuccessfulUploadDisconnect = 0;
	m_iSessionSuccessfulUploadLimit = 0;
	m_iSessionSuccessfulUploadOther = 0;
	m_iSessionFailedUploadCancel = 0;
	m_iSessionFailedUploadTimeout = 0;
	m_iSessionFailedUploadDisconnect = 0;
	m_iSessionFailedUploadLimit = 0;
	m_iSessionFailedUploadUnreachable = 0;
	m_iSessionFailedUploadRejected = 0;
	m_iSessionFailedUploadOther = 0;
	// Configuration
	m_iTaskPriority = TP_Normal;
	m_iSmallFilePush = SFP_Nonlinear;
	m_bMaxUpSlotsEnabled = false;
	m_bMinUpDatarateEnabled = false;
	m_iMaxUpSlots = 10;
	m_iMinUpDatarate = 3072;
	m_iUpSlotInterval = 0;
	// Features
	m_bPaybackQueue = true;
	m_bFastQueueRotation = true;
	m_bFreshmanCredits = true;
	m_bDownloadHidden = true;
	m_bRarePartCredits = true;
}

void CNetF::Init()
{
	LoadPreferences();
}

void CNetF::SavePreferences() const
{
	USES_CONVERSION;
	CString buffer;
	TCHAR* fullpath = new TCHAR[_tcslen(thePrefs.GetConfigDir())+MAX_PATH]; // i_a
	_stprintf(fullpath,_T("%sNetF.ini"),thePrefs.GetConfigDir());
	
	CIni ini( fullpath, _T("NetF") );
	delete[] fullpath;
	fullpath=NULL;
	
	ini.WriteInt(_T("TaskPriority"), m_iTaskPriority);

	ini.WriteInt(_T("SmallFilePush"), m_iSmallFilePush, _T("CreditSystem"));

	ini.WriteBool(_T("MaxUpSlotsEnabled"), m_bMaxUpSlotsEnabled, _T("UploadManagement"));
	ini.WriteInt(_T("MaxUpSlots"), m_iMaxUpSlots);
	ini.WriteBool(_T("MinUpDatarateEnabled"), m_bMinUpDatarateEnabled);
	ini.WriteInt(_T("MinUpDatarate"), m_iMinUpDatarate);
	ini.WriteInt(_T("MinUpSlotInterval"), m_iUpSlotInterval);

	ini.WriteBool(_T("PaybackQueue"), m_bPaybackQueue, _T("Features"));
	ini.WriteBool(_T("FastQueueRotation"), m_bFastQueueRotation);
	ini.WriteBool(_T("FreshmanCredits"), m_bFreshmanCredits);
	ini.WriteBool(_T("DownloadHidden"), m_bDownloadHidden);
	ini.WriteBool(_T("RarePartCredits"), m_bRarePartCredits);
}

void CNetF::LoadPreferences()
{
	USES_CONVERSION;
	CString buffer;
	TCHAR* fullpath = new TCHAR[_tcslen(thePrefs.GetConfigDir())+MAX_PATH]; // i_a
	_stprintf(fullpath,_T("%sNetF.ini"),thePrefs.GetConfigDir());
	
	CIni ini( fullpath, _T("NetF") );
	delete[] fullpath;
	fullpath=NULL;

	m_iTaskPriority = (ETaskPriority) ini.GetInt(_T("TaskPriority"), TP_Normal);
	UpdateTaskPriority();

	m_iSmallFilePush = (ESmallFilePush) ini.GetInt(_T("SmallFilePush"), SFP_Nonlinear, _T("CreditSystem"));

	m_bMaxUpSlotsEnabled = ini.GetBool(_T("MaxUpSlotsEnabled"), false, _T("UploadManagement"));
	m_iMaxUpSlots = ini.GetInt(_T("MaxUpSlots"), 10);
	m_bMinUpDatarateEnabled = ini.GetBool(_T("MinUpDatarateEnabled"), false);
	m_iMinUpDatarate = ini.GetInt(_T("MinUpDatarate"), 3072);
	m_iUpSlotInterval = ini.GetInt(_T("MinUpSlotInterval"), 0);

	m_bPaybackQueue = ini.GetBool(_T("PaybackQueue"), true, _T("Features"));
	m_bFastQueueRotation = ini.GetBool(_T("FastQueueRotation"), true);
	m_bFreshmanCredits = ini.GetBool(_T("FreshmanCredits"), true);
	m_bDownloadHidden = ini.GetBool(_T("DownloadHidden"), true);
	m_bRarePartCredits = ini.GetBool(_T("RarePartCredits"), true);
}

void CNetF::UpdateTaskPriority()
{
	switch(m_iTaskPriority)
	{
	case TP_Low:
		SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS);
		break;
	case TP_BelowNormal:
		SetPriorityClass(GetCurrentProcess(), BELOW_NORMAL_PRIORITY_CLASS);
		break;
	case TP_AboveNormal:
		SetPriorityClass(GetCurrentProcess(), ABOVE_NORMAL_PRIORITY_CLASS);
		break;
	case TP_High:
		SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
		break;
	case TP_Normal:
	default:
		SetPriorityClass(GetCurrentProcess(), NORMAL_PRIORITY_CLASS);
		break;
	}
}