//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "NetF.h"
#include <math.h>

uint32 CUpDownClient::GetScore(bool sysvalue, bool isdownloading, bool onlybasevalue, bool paybackslot) const
{
	if (!m_pszUsername)
		return 0;

	if (credits == 0){
		ASSERT ( IsKindOf(RUNTIME_CLASS(CUrlClient)) );
		return 0;
	}
	CKnownFile* currequpfile = (m_pUpFileStatus ? m_pUpFileStatus->GetFile() : NULL);
	if(!currequpfile)
		return 0;
	
	// bad clients (see note in function)
	if (credits->GetCurrentIdentState(GetIP()) == IS_IDBADGUY)
		return 0;
	// friend slot
	if (IsFriend() && GetFriendSlot() && !(HasLowID() || m_bPassiveConnectOnly))
		return 0x0FFFFFFF;

	if (IsBanned() || m_bGPLEvildoer)
		return 0;

	if (sysvalue && (HasLowID() || m_bPassiveConnectOnly) && !(socket && socket->IsConnected())){
		return 0;
	}

	if (!isdownloading && !onlybasevalue){
		if (sysvalue && (HasLowID() || m_bPassiveConnectOnly) && !(socket && socket->IsConnected()) ){
			if (!theApp.serverconnect->IsConnected() || (theApp.serverconnect->IsLowID()) || theApp.listensocket->TooManySockets()) //This may have to change when I add firewall support to Kad
				return 0;
		}
	}

	// MOD BEGIN netfinity: Setup rules for credits
	bool	enableupcredits = (credits->GetDownloadedTotal() > 0); // Allow credits as soon client uploaded to us
	double	sessionratio; // This value should be as high as possible
	if(theStats.sessionReceivedBytes > 0)
		sessionratio = ((double) theStats.sessionSentBytes) / ((double) theStats.sessionReceivedBytes);
	else 
		sessionratio = 1.0;
	// MOD END netfinity

	// MOD BEGIN netfinity: Client bonus we always give
	double	clientmodif = 1.0;
	// Uploading clients are always prioritized
	if(GetTransferredDown() >= GetTransferredUp() + PARTSIZE) clientmodif *= 1.1;
	// Best roundtrip client (faulty)
	//if(GetAskedCountDown() > 3 && GetAskedCount() > 3 && GetTransferedUp())
	//{
	//	double	ratioperask = (GetTransferedDown() / GetAskedCountDown());
	//	ratioperask /= (GetTransferedUp() / GetAskedCount());
	//	ratioperask = sqrt(ratioperask) / 10.0;
	//	if(ratioperask < 0.95) ratioperask = 0.95;
	//	if(ratioperask < 1.05) ratioperask = 1.05;
	//	clientmodif *= ratioperask;
	//	if(onlybasevalue && !sysvalue)
	//		AddDebugLogLine(false,"Ratio/ask modifier: %f",ratioperask);
	//}
	// Disadvantage clients who asks to often (creates high load)
	if(GetAskedCount() > 5)
	{
		double askfactor = GetWaitStartTime().Elapsed(GetLastUpRequest()) / ((GetAskedCount() - 1) * MIN2MS(20));
		if(askfactor > 1.1) askfactor = 1.1;
		clientmodif *= askfactor;
	}
	// Reduce number of slots given to saturated clients
	if(MOD_GetEstimatedUploadRate() < 1024)
		clientmodif /= 2.0;
	// MOD END netfinity

	// MOD BEGIN netfinity: Payback slot score calculation
	if(paybackslot)
	{
		CBytes	payback = credits->GetDownloadedTotal() - credits->GetUploadedTotal();
		// Try payback this session atleast, but then with a lower score
		payback = max(payback, (GetTransferredDown() >= EMBLOCKSIZE ? (GetTransferredDown() - GetTransferredUp()) / 2 : 0));
		uint32	score = 0;
		if(payback <= 0 || !enableupcredits || sessionratio <= 1.1)
		{	// Insufficent upload from client or our session ratio is to low
			score = rand(); // Make a random pick instead
		}
		else
		{
			// check the client ident status
			if ( ( credits->GetCurrentIdentState(GetIP()) == IS_IDFAILED || credits->GetCurrentIdentState(GetIP()) == IS_IDBADGUY || credits->GetCurrentIdentState(GetIP()) == IS_IDNEEDED) && theApp.clientcredits->CryptoAvailable() )
			{	// bad guy - no payback for you
				score = rand(); // Make a random pick instead
			}
			else
			{
				// Payback as long the other client doesn't owe us anything 
				score = payback / CBytes(256 * 1024);	// Reduce to fit uint32
				if (!isdownloading)
					score = score + (GetWaitStartTime().Elapsed() + m_dwUploadTimeBonus) / MIN2MS(1); // Add waiting time to calculation
				else
					score = score + (GetWaitStartTime().Elapsed(m_dwUploadTime) + m_dwUploadTimeBonus) / MIN2MS(1); // Add waiting time to calculation
				score = 1000000 + score * clientmodif;
			}
		}
		return ((uint32) score); 
	}
	// MOD END netfinity

	// TODO coded by tecxx & herbert, one yet unsolved problem here:
	// sometimes a client asks for 2 files and there is no way to decide, which file the 
	// client finally gets. so it could happen that he is queued first because of a 
	// high prio file, but then asks for something completely different.
	int filepriority = 10; // standard
	switch(currequpfile->GetUpPriority()){
		case PR_VERYHIGH:
			filepriority = 18;
			break;
		case PR_HIGH: 
			filepriority = 9; 
			break; 
		case PR_LOW: 
			filepriority = 6; 
			break; 
		case PR_VERYLOW:
			filepriority = 2;
			break;
		case PR_NORMAL: 
			default: 
			filepriority = 7; 
		break; 
	}
	// MOD BEGIN netfinity: Override filepriorities for part files
	//	Disabled as of NetF 0.2g
	/*if(currequpfile->IsPartFile())
	{
		switch(static_cast<CPartFile*>(currequpfile)->GetDownPriority())
		{
		case PR_HIGH:
			if(filepriority < 11) filepriority = 11; 
			break;
		case PR_NORMAL:
			if(filepriority < 8) filepriority = 8; 
			break;
		case PR_LOW: 
			// Keep it as it is! 
			break; 
		}
	}*/
	// MOD END netfinity
	// MOD BEGIN netfinity: Boost files not yet uploaded
	/*double filefactor = 10.0;
	if (currequpfile->statistic.GetAllTimeTransferred() > 0)
		filefactor = (double) currequpfile->GetFileSize() / (double) currequpfile->statistic.GetAllTimeTransferred();
	if (filefactor > 10.0)
		filefactor = 10.0;
	if (filefactor < 1.0)
		filefactor = 1.0;
	if (currequpfile->IsPartFile() && filefactor > 2.0)
		filefactor = 2.0;*/
	// MOD END netfinity

	// calculate score, based on waitingtime and other factors
	float fBaseValue;
	if (onlybasevalue)
		fBaseValue = 100;
	else if (!isdownloading)
		fBaseValue = (float) (GetWaitStartTime().Elapsed() + m_dwUploadTimeBonus).GetValue() / 1000;
	else{
		// we dont want one client to download forever
		// the first 15 min downloadtime counts as 15 min waitingtime and you get a 15 min bonus while you are in the first 15 min :)
		// (to avoid 20 sec downloads) after this the score won't raise anymore 
		fBaseValue = (float) (GetWaitStartTime().Elapsed(m_dwUploadTime) + m_dwUploadTimeBonus).GetValue();
		ASSERT ( GetWaitStartTime().Elapsed(m_dwUploadTime) >= 0 ); //oct 28, 02: changed this from "> 0" to ">= 0"
		fBaseValue += (float)(m_dwUploadTime.Elapsed() > 900000)? 900000:1800000;
		fBaseValue /= 1000;
	}
	if(thePrefs.UseCreditSystem())
	{
		float modif = credits->GetScoreRatio(GetIP());
		if (IsFriend() && modif < 2.0) // Friends don't get any penalty
			modif = 2.0;
		fBaseValue *= modif;
	}
	if (!onlybasevalue)
		fBaseValue *= (float(filepriority)/10.0f) /* * filefactor*/;
	if( (IsEmuleClient() || this->GetClientSoft() < 10) && m_byEmuleVersion <= 0x19 )
		fBaseValue *= 0.5f;

	// MOD BEGIN netfinity: Push small files
	if(theNetF.GetSmallFilePush() != SFP_None && currequpfile->GetFileSize() > 0 && currequpfile->GetFileSize() < PARTSIZE)
	{
		float modif = ((double) PARTSIZE) / ((double) currequpfile->GetFileSize());
		if(theNetF.GetSmallFilePush() == SFP_Nonlinear)
			modif = sqrt(modif);
		if(modif < 1.0) modif = 1.0;
		if(modif > 1000.0) modif = 1000.0;
		fBaseValue *= modif;
	}
	// MOD END netfinity

	// MOD BEGIN netfinity: Check for possible leecher!
	//switch(GetDownloadState())
	//{
	//case DS_DOWNLOADING:
	//case DS_ONQUEUE:
	//case DS_REMOTEQUEUEFULL:
	//	// Apparently this guy has something we need so it's very strange
	//	//  if he wouldn't be able to upload some.
	//	if((GetTransferedUp() - GetTransferedDown()) > (5 * PARTSIZE))
	//	{
	//		fBaseValue *= 0.75; // Reduce his credit's until the situation has rectified
	//	}
	//	break;
	//default:
	//	break;	// We don't need anything from this guy
	//}
	// MOD END netfinity

	fBaseValue *= clientmodif;
	return (uint32)fBaseValue;
}

// MOD BEGIN netfinity:
//  Code by STORMIT, Pawcio and more...
/** 
 * @brief Checks if it is next requested block from another chunk of the actual file or from another file 
 *
 * @return 
 *   true : Next requested block is from another different chunk or file than last downloaded block 
 *   false: Next requested block is from same chunk that last downloaded block 
 */
bool CUpDownClient::MOD_IsDifferentPartBlock()
{ 
	Requested_Block_Struct* lastBlock;
	Requested_Block_Struct* currBlock;
	uint32 lastDone = 0;
	uint32 currRequested = 0;
	
	bool different = false;
	
	try {
		// Check if we have good lists and proceed to check for different chunks
		bool reqnotempty=!m_BlockRequests_queue.IsEmpty();
		bool donenotempty=!m_DoneBlocks_list.IsEmpty();
		if ( reqnotempty && donenotempty)
		{
			// Calculate corresponding parts to blocks
			lastBlock = (Requested_Block_Struct*)m_DoneBlocks_list.GetTail();
			currBlock = (Requested_Block_Struct*)m_BlockRequests_queue.GetHead(); 
			lastDone = lastBlock->StartOffset / PARTSIZE;
			currRequested = currBlock->StartOffset / PARTSIZE; 
             
			// Test is we are asking same file and same part
			if ( lastDone != currRequested)
			{ 
				different = true;
				//theApp.emuleDlg->QueueDebugLogLine(false, "%s: Upload session ended due to new chunk.", this->GetUserName());
			}
			if (md4cmp(lastBlock->FileID, currBlock->FileID) != 0)
			{ 
				different = true;
				//theApp.emuledlg->QueueDebugLogLine(false, "%s: Upload session ended due to different file.", this->GetUserName());
			}
		} 
   	}
   	catch(...)
   	{ 
			AddDebugLogLine(DLP_HIGH, false, _T("CUpDownClient::MOD_IsDifferentPartBlock - An unknown exception occurred"));
      		different = true; 
   	} 
//	AddDebugLogLine(false, "Debug: User %s, last_done_part (%u) %s (%u) next_requested_part, sent %u Kbs.", GetUserName(), last_done_part, different_part? "!=": "==", next_requested_part, this->GetTransferedUp() / 1024); 

	return different; 
}
// MOD END netfinity

CBytes CUpDownClient::MOD_GetMaxTransfer(bool checkLastTransfer)
{
	Requested_Block_Struct* block = NULL;
	uint32 requested = 0;
	uint16 partSrcCount = 0;

	try {
	if(!m_BlockRequests_queue.IsEmpty() && !checkLastTransfer)
	{
		block = (Requested_Block_Struct*)m_BlockRequests_queue.GetHead();
	}
	else if (!m_DoneBlocks_list.IsEmpty())
	{
		block = (Requested_Block_Struct*)m_DoneBlocks_list.GetTail(); 
	}
	if(block)
	{
		requested = block->StartOffset / PARTSIZE;
		if(m_pUpFileStatus) partSrcCount = m_pUpFileStatus->GetFile()->MOD_GetPartSourceCount(requested);
	}

	// This is to prevent common parts to get to much focus
	const uint16 partSrcTreshold = 20;
	if(partSrcCount < partSrcTreshold) partSrcCount = partSrcTreshold;
	return (partSrcTreshold * PARTSIZE) / partSrcCount + (2 * 3 * EMBLOCKSIZE);
	}
	catch(...)
   	{ 
			AddDebugLogLine(DLP_HIGH, false, _T("%s; An unknown exception occurred"), _T(__FUNCTION__));
   	}
	return PARTSIZE;
}

bool CUpDownClient::MOD_IsPartialSource()
{
	return (m_pUpFileStatus == m_pDownFileStatus);
}

CBytesPerSecond CUpDownClient::MOD_GetEstimatedUploadRate(void) const
{
	CBytesPerSecond uploadrate = 3072; // default client datarate

	if(m_dwCumulativeUploadTime >= SEC(60)) // At least one minute of upload before calculating datarate
	{
		 uploadrate = m_qwCumulativeUploadBytes / m_dwCumulativeUploadTime;
	}
	if(GetUploadState() == US_UPLOADING)
	{
		if(GetUpStartTimeDelay() >= SEC2MS(60)) // At least one minute of upload before calculating datarate
		{
			uploadrate = GetSessionUp() / GetUpStartTimeDelay();
		}
	}
	return uploadrate;
}

void CUpDownClient::ProcessExtendedInfo(CSafeMemFile* data, CKnownFile* tempreqfile)
{
	CClientFileStatus*	pFileStatus = m_FileStatusList.AddFile(tempreqfile); // netfinity: New style part status records
	if (pFileStatus == NULL)
		throw CString(_T("ProcessExtendedInfo; pFileStatus == NULL"));
	if ( GetExtendedRequestsVersion() == 0 )
		return;
	else if (data->GetLength() - data->GetPosition() == 0)
	{
		return;
	}
	uint16 nED2KUpPartCount = data->ReadUInt16();
	bool bPartsNeeded = false;
	if (!nED2KUpPartCount)
	{
		pFileStatus->SetCompleteFile();
		bPartsNeeded = true;
	}
	else
	{
		if (tempreqfile->GetED2KPartCount() != nED2KUpPartCount)
		{
			//We already checked if we are talking about the same file.. So if we get here, something really strange happened!
			//m_nUpPartCount = 0;
			throw CString(_T(__FUNCTION__) _T("; ed2k part count mismatch"));
			return;
		}
		uint16 partCount = tempreqfile->GetPartCount();
		//fileStatus->SetPartCount(partCount);
		uint16 done = 0;
		while (done != partCount)
		{
			uint8 toread = data->ReadUInt8();
			for (sint32 i = 0;i != 8;i++){
				if((toread >> i) & 1)
				{
					if (tempreqfile->IsPartFile() && !(static_cast<CPartFile*>(tempreqfile)->IsComplete(done*PARTSIZE, ((done+1)*PARTSIZE)-1, false)))
						bPartsNeeded = true;
					pFileStatus->SetCompletePart(done);
				}
				else
					pFileStatus->ClearCompletePart(done);
				done++;
				if (done == partCount)
					break;
			}
		}
		if (GetExtendedRequestsVersion() > 1 && (data->GetLength() - data->GetPosition()) >= 2)
		{
			uint16 nCompleteCountLast = pFileStatus->GetCompleteSources();
			uint16 nCompleteCountNew = data->ReadUInt16();
			pFileStatus->SetCompleteSources(nCompleteCountNew);
			if (nCompleteCountLast != nCompleteCountNew)
			{
				tempreqfile->UpdatePartsInfo();
			}
		}
	}
	if(bPartsNeeded)
		pFileStatus->SetHasNeededParts();
	// netfinity: Put us back in queue when seeing new parts from a passive NNP source
	//	This allow us to save some overhead by reducing the need to ask NNP sources
	if (bPartsNeeded && m_pDownFileStatus == pFileStatus && GetDownloadState() == DS_NONEEDEDPARTS)
	{
		if (socket && socket->IsConnected())
		{
			if (pFileStatus->GetLastDownAsked().IsTickSet() && pFileStatus->GetLastDownAsked().Elapsed() < MIN_REQUESTTIME)
			{
				SetDownloadState(DS_ONQUEUE);
			}
			else if (static_cast<CPartFile*>(tempreqfile)->hashsetneeded)
			{
				if (socket)
				{
					if (thePrefs.GetDebugClientTCPLevel() > 0)
						DebugSend("OP__HashSetRequest", this);
					Packet* packet = new Packet(OP_HASHSETREQUEST,16);
					md4cpy(packet->pBuffer,tempreqfile->GetFileHash());
					theStats.AddUpDataOverheadFileRequest(packet->size);
					socket->SendPacket(packet, true, true);
					SetDownloadState(DS_REQHASHSET);
					m_fHashsetRequesting = 1;
					static_cast<CPartFile*>(tempreqfile)->hashsetneeded = false;
				}
				else
					ASSERT(0);
			}
			else
			{
				SendStartupLoadReq();
			}
		}
		else
			SetDownloadState(DS_ONQUEUE);
	}
	theApp.emuledlg->transferwnd->queuelistctrl.RefreshClient(this);
}

void
CUpDownClient::MOD_SendOutOfPartReqsAndFileNotFound(void)
{
	if (thePrefs.GetDebugClientTCPLevel() > 0)
		DebugSend("OP__OutOfPartReqs", this);
	Packet* pOOPPacket = new Packet(OP_OUTOFPARTREQS, 0);
	theStats.AddUpDataOverheadFileRequest(pOOPPacket->size);
	socket->SendPacket(pOOPPacket, true, true);
	m_fSentOutOfPartReqs = 1;
	// send file request no such file packet (0x48)
	if (thePrefs.GetDebugClientTCPLevel() > 0)
		DebugSend("OP__FileReqAnsNoFil", this);
	Packet* pFNFPacket = new Packet(OP_FILEREQANSNOFIL, 16);
	md4cpy(pFNFPacket->pBuffer, m_pUpFileStatus->GetFile()->GetFileHash());
	theStats.AddUpDataOverheadFileRequest(pFNFPacket->size);
	socket->SendPacket(pFNFPacket, true);
}
