//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "PartFile.h"
#include "NetF.h"

bool CUploadQueue::AddUpNextClient(LPCTSTR pszReason, CUpDownClient* directadd){
	POSITION toadd = 0;
	POSITION toaddlow = 0;
	uint32	bestscore = 0;
	uint32  bestlowscore = 0;
	CUpDownClient* newclient;
	bool	allowAddNextConnect = true;
	if (m_tickLastGiveOnReconnect.IsTickSet() && m_tickLastGiveOnReconnect.Elapsed() < SEC2MS(10))
		allowAddNextConnect = false;
	// MOD BEGIN netfinity: Payback slot
	bool	paybackslot = false;
	m_slotcounter++;
	if(m_slotcounter == 1 && theNetF.m_bPaybackQueue) paybackslot = true;
	if(m_slotcounter >= 3) m_slotcounter = 0;

	// select next client or use given client
	if (!directadd)
	{
		POSITION pos1, pos2;
		for (pos1 = waitinglist.GetHeadPosition();( pos2 = pos1 ) != NULL;)
		{
			waitinglist.GetNext(pos1);
			CUpDownClient* cur_client =	waitinglist.GetAt(pos2);
			// clear dead clients
			ASSERT ( cur_client->GetLastUpRequest().IsTickSet() );
			if ((cur_client->GetLastUpRequest().IsTickSet() && cur_client->GetLastUpRequest().Elapsed() > MAX_PURGEQUEUETIME) || !theApp.sharedfiles->GetFileByID(cur_client->GetUploadFileID()) )
			{
				cur_client->ClearWaitStartTime();
				RemoveFromWaitingQueue(pos2,true);
				continue;
			}
			// netfinity: check if client is ready to receive an upload slot (Inspired by Xman)
			if (!cur_client->m_bAllowUploadSlot)
				continue;
			// netfinity: filter clients not needing any parts
			if (cur_client->m_pUpFileStatus && cur_client->m_pUpFileStatus->GetPartsNeededByClient() == 0)
				continue;
			// finished clearing
			uint32 cur_score = cur_client->GetScore(true,false,false,paybackslot);
			// calculate best score
			if ( cur_score > bestscore)
			{
				bestscore = cur_score;
				toadd = pos2;
			} 
			else if (cur_score == 0 && allowAddNextConnect) // netfinity: Avoid unneccessary overhead
			{
				cur_score = cur_client->GetScore(false,false,false,paybackslot);
				// calculate best score
				if ((cur_score > bestlowscore) && !cur_client->m_bAddNextConnect)
				{
					bestlowscore = cur_score;
					toaddlow = pos2;
				}
			}
		}
		
		if (bestlowscore > bestscore && allowAddNextConnect)
		{
			m_tickLastGiveOnReconnect.SetCurrentTick();
			newclient = waitinglist.GetAt(toaddlow);
			newclient->m_bAddNextConnect = true;
			// netfinity: Remember why the slot was given!
			newclient->m_bSlotGivenIsPayback = (paybackslot && bestlowscore > 900000);
			newclient->m_dwSlotGivenScore = bestlowscore;
			newclient->m_pSlotGivenForFile = newclient->m_pUpFileStatus;
		}
		if (!toadd)
			return false;
		newclient = waitinglist.GetAt(toadd);
		// netfinity: Remember why the slot was given!
		newclient->m_bSlotGivenIsPayback = (paybackslot && bestscore > 900000);
		newclient->m_dwSlotGivenScore = bestscore;
		newclient->m_pSlotGivenForFile = newclient->m_pUpFileStatus;
		/*
		if(paybackslot)
		{
			if(bestscore < 900000)
				AddLogLine(true,_T("Added random slot: %s"), newclient->GetUserName());
			else
				AddLogLine(true,_T("Added payback slot: %s"), newclient->GetUserName());
		}
		else
			AddLogLine(true,_T("Added normal slot: %s"), newclient->GetUserName());
		*/
		RemoveFromWaitingQueue(toadd, true);
		theApp.emuledlg->transferwnd->ShowQueueCount(waitinglist.GetCount());
	}
	else 
	{
		newclient = directadd;
		// netfinity: Remember why the slot was given!
		if (newclient->m_pSlotGivenForFile == NULL)
		{
			newclient->m_bSlotGivenIsPayback = false;
			newclient->m_dwSlotGivenScore = newclient->GetScore(false,false,false,false);
			newclient->m_pSlotGivenForFile = newclient->m_pUpFileStatus;
		}

		//AddLogLine(true,_T("Added direct slot: %s"), newclient->GetUserName());
		/*if (!IsDownloading(newclient)){
            if (newclient->HasLowID())
                AddLogLine(true,"DirectAdd:  LowID: %s", newclient->GetUserName());
			else
                AddLogLine(true,"DirectAdd:  HighID: %s", newclient->GetUserName());
		}*/
	}
	// MOD END netfinity
    if(newclient == NULL) {
        return false;
    }
	newclient->m_dwRemovePending.UnsetTick(); // netfinity: We don't want the slot removed instantly after it has been added

	if (!thePrefs.TransferFullChunks() && !theNetF.m_bFastQueueRotation) // netfinity: Session length is not affected by score when Fast Queue Rotation 
		UpdateMaxClientScore(); // refresh score caching, now that the highest score is removed

	if (IsDownloading(newclient))
	{
		return false;
	}
	// tell the client that we are now ready to upload
	if (!newclient->socket || !newclient->socket->IsConnected())
	{
		newclient->SetUploadState(US_CONNECTING);
		if (!newclient->TryToConnect(true))
			return false;
	}
	else
	{
		if (thePrefs.GetDebugClientTCPLevel() > 0)
			DebugSend("OP__AcceptUploadReq", newclient);
		Packet* packet = new Packet(OP_ACCEPTUPLOADREQ,0);
		theStats.AddUpDataOverheadFileRequest(packet->size);
		newclient->SetUploadState(US_UPLOADING); // netfinity: Swapped order just to be on the safe side
		newclient->socket->SendPacket(packet,true);
	}
	newclient->SetUpStartTime();
	newclient->ResetSessionUp();

    InsertInUploadingList(newclient);

    m_nLastStartUpload.SetCurrentTick();
	
	// statistic
	CKnownFile* reqfile = theApp.sharedfiles->GetFileByID((uchar*)newclient->GetUploadFileID());
	if (reqfile){
		reqfile->statistic.AddAccepted();
	}
		
	theApp.emuledlg->transferwnd->uploadlistctrl.AddClient(newclient);

	return true;
}

bool CUploadQueue::RemoveFromUploadQueue(CUpDownClient* client, LPCTSTR pszReason, bool updatewindow, bool earlyabort, EReason nReason){
    bool result = false;
    uint32 slotCounter = 1;
	// MOD BEGIN netfinity: Delayed accept upload 
	if (client)
		client->MOD_ClearAcceptUploadPending();
	// MOD END netfinity
	for (POSITION pos = uploadinglist.GetHeadPosition();pos != 0;){
        POSITION curPos = pos;
        CUpDownClient* curClient = uploadinglist.GetNext(pos);
		if (client == curClient){
			if (updatewindow)
				theApp.emuledlg->transferwnd->uploadlistctrl.RemoveClient(client);

			if (thePrefs.GetLogUlDlEvents())
				AddDebugLogLine(DLP_VERYLOW, true,_T("---- %s: Removing client from upload list. Reason: %s ----"), client->DbgGetClientInfo(), pszReason==NULL ? _T("") : pszReason);
            uploadinglist.RemoveAt(curPos);

            bool removed = theApp.uploadBandwidthThrottler->RemoveFromStandardList(client->socket);
            bool pcRemoved = theApp.uploadBandwidthThrottler->RemoveFromStandardList((CClientReqSocket*)client->m_pPCUpSocket);

            //if(thePrefs.GetLogUlDlEvents() && !(removed || pcRemoved)) {
            //    AddDebugLogLine(false, _T("UploadQueue: Didn't find socket to delete. Adress: 0x%x"), client->socket);
            //}

			if(client->GetSessionUp() > 0) {
				++successfullupcount;
				totaluploadtime += client->GetUpStartTimeDelay(); // /1000;
				// MOD BEGIN netfinity: Upload bandwidth estimation
				client->m_qwCumulativeUploadBytes += client->GetSessionUp();
				client->m_dwCumulativeUploadTime += client->GetUpStartTimeDelay();
				// MOD END netfinity
            } else if(earlyabort == false)
				++failedupcount;

	        CKnownFile* requestedFile = theApp.sharedfiles->GetFileByID(client->GetUploadFileID());

	        if(requestedFile != NULL) {
		        requestedFile->UpdatePartsInfo();

				// MOD BEGIN netfinity:
				if (client->m_bSlotGivenIsPayback)
				{
					// Keep waiting time
					client->m_dwUploadTimeBonus = client->GetWaitStartTime().Elapsed(client->m_dwUploadTime) + client->m_dwUploadTimeBonus;
				}
				else
				{
					// Incomplete session bonus
					float modif = client->GetSessionUp() / client->MOD_GetMaxTransfer(true);
					if(modif < 0.1f) modif = 0.1f;
					if(modif > 1.0f) modif = 1.0f;
					client->m_dwUploadTimeBonus = (client->GetWaitStartTime().Elapsed(client->m_dwUploadTime) + client->m_dwUploadTimeBonus * 0.9f) * (1.0f - modif);
				}
				// MOD END netfinity
            }
			client->m_dwRemovePending.UnsetTick(); // netfinity: Just to be on the safe side!
			// MOD BEGIN netfinity: Forget why the slot was given!
			client->m_bSlotGivenIsPayback = false;
			client->m_dwSlotGivenScore = 0;
			client->m_pSlotGivenForFile = NULL;
			// MOD END netfinity
			theApp.clientlist->AddTrackClient(client); // Keep track of this client
			client->SetUploadState(US_NONE, nReason);
			client->ClearUploadBlockRequests();

            m_iHighestNumberOfFullyActivatedSlotsSinceLastCall = 0;

			result = true;
        } else {
            curClient->SetSlotNumber(slotCounter);
            slotCounter++;
        }
	}
	return result;
}

bool CUploadQueue::CheckForTimeOver(CUpDownClient* client){
	// MOD BEGIN netfinity:
	bool		canRemove = true;
	bool		timeOver = false;
	CEMSocket*	uploadSocket = NULL;

	try
	{
	uploadSocket = client->GetFileUploadSocket(); // Allow this to work for PeerCache too

	if (client->GetFriendSlot()) // Don't drop friendslots
		return false;

	if (waitinglist.IsEmpty() && AcceptNewClient(true) || ForceNewClient())
		canRemove = false; // No one on queue or more slots required, so keep uploading

	if (client->m_dwRemovePending.IsTickSet())
	{
		timeOver = true;
		if (client->m_dwRemovePending.Elapsed() > MIN2MS(5)) // Prevent slot to be in ending state forever (5 minutes)
			return canRemove; // Force remove if waiting queue is not empty
	}

	if (timeOver) // Early return if remove is already pending
	{
		if(uploadSocket != NULL)
		{
			uploadSocket->m_eThrottleState = THROTTLE_ENDING; // Ask download throttler to finish off this client
			return (canRemove && uploadSocket->m_bNotNeeded); // Remove if throttler doesn't need the slot anymore
		}
		else return canRemove;
	}

	// Check if client is downloading another file than the slot was given for
	if (client->m_pSlotGivenForFile != client->m_pUpFileStatus)
	{
		if (!client->m_bSlotGivenIsPayback && client->m_dwSlotGivenScore < client->GetScore(false, false, false, false))
		{
			if (thePrefs.GetLogUlDlEvents())
				AddDebugLogLine(DLP_LOW, false, _T("%s: Upload session ended due to other file than slot given for. (%s uploaded)"), client->GetUserName(), CastItoXBytes(client->GetQueueSessionPayloadUp()));
			if (canRemove)
				return true; // Forced dropping if other client can take it's place, in order to deal with leechers
			else
				timeOver = true;
		}
		else
			client->m_pSlotGivenForFile = client->m_pUpFileStatus; // Let client swap to this file
	}

	// Check if the there is anything left to payback, to prevent abuse
	if (client->m_bSlotGivenIsPayback && client->credits && client->GetQueueSessionPayloadUp() >= EMBLOCKSIZE)
	{
		// Check if payback exceed 150% of the amount to payback, but allow it to exceed atleast 180kB
		if(
			(3 * client->credits->GetDownloadedTotal() < 2 * client->credits->GetUploadedTotal()) // Total amount exceeded
			&& (3 * client->GetTransferredDown() < 2 * client->GetTransferredUp()) // Session amount (150%) exceeded
			&& (client->GetTransferredDown() + EMBLOCKSIZE < client->GetTransferredUp()) // Session amount (+180k) exceeded
		   )
		{
			if (thePrefs.GetLogUlDlEvents())
				AddDebugLogLine(DLP_LOW, false, _T("%s: Upload session ended due to payback exceed 150%%. (%s uploaded)"), client->GetUserName(), CastItoXBytes(client->GetQueueSessionPayloadUp()));
			timeOver = true;
		}
	}

	// Keep upload slots for atleast 5 mins
	if (client->GetUpStartTimeDelay() < MIN2MS(5) && !timeOver)
		return false;

	if (!theNetF.m_bFastQueueRotation)
	{
		if (!thePrefs.TransferFullChunks())
		{
			if( client->GetUpStartTimeDelay() > SESSIONMAXTIME){ // Try to keep the clients from downloading for ever
				if (thePrefs.GetLogUlDlEvents())
					AddDebugLogLine(DLP_LOW, false, _T("%s: Upload session ended due to max time %s."), client->GetUserName(), CastSecondsToHM(SESSIONMAXTIME/1000));
				timeOver = true;
			}
			else
			{
				// Cache current client score
				const uint32 score = client->GetScore(true, true);

				// Check if another client has a bigger score
				if (score < GetMaxClientScore() && m_dwRemovedClientByScore.Elapsed() > 0) {
					if (thePrefs.GetLogUlDlEvents())
						AddDebugLogLine(DLP_VERYLOW, false, _T("%s: Upload session ended due to score."), client->GetUserName());
					//Set timer to prevent to many uploadslot getting kick do to score.
					//Upload slots are delayed by a min of 1 sec and the maxscore is reset every 5 sec.
					//So, I choose 6 secs to make sure the maxscore it updated before doing this again.
					m_dwRemovedClientByScore = CTick::GetCurrentTick() + SEC2MS(6);
					timeOver = true;
				}
			}
		}
		else
		{
			// Allow the client to download a specified amount per session	
			if( client->GetQueueSessionPayloadUp() > SESSIONMAXTRANS ){
				if (thePrefs.GetLogUlDlEvents())
					AddDebugLogLine(DLP_DEFAULT, false, _T("%s: Upload session ended due to max transfered amount. %s"), client->GetUserName(), CastItoXBytes(SESSIONMAXTRANS));
				timeOver = true;
			}
		}
	}
	else
	{
		// Keep transfers within chunk but don't end if transfer is small
		if (client->MOD_IsDifferentPartBlock() && client->GetQueueSessionPayloadUp() >= (6 * EMBLOCKSIZE))
		{
			if (thePrefs.GetLogUlDlEvents())
				AddDebugLogLine(DLP_DEFAULT, false, _T("%s: Upload session ended due to not same part or file. %s"), client->GetUserName(), CastItoXBytes(client->GetQueueSessionPayloadUp()));
			timeOver = true;
		}
		// Allow the client to download a specified amount per session
		CBytes	maxTransfer = client->MOD_GetMaxTransfer(false);
		if( client->GetQueueSessionPayloadUp() > maxTransfer){
			if (thePrefs.GetLogUlDlEvents())
				AddDebugLogLine(DLP_DEFAULT, false, _T("%s: Upload session ended due to max transfered amount. %s"), client->GetUserName(), CastItoXBytes(maxTransfer));
			timeOver = true;
		}
		// Check for client saturation
		uint16	completeSrcCount = 1;
		if(client->m_pUpFileStatus) 
			completeSrcCount = client->m_pUpFileStatus->GetFile()->m_nCompleteSourcesCount;
		if (client->GetUpStartTimeDelay() > MIN2MS(5) && completeSrcCount > RARE_FILE) // if uploaded more than 5 minutes and file is not rare
		{	
			CBytesPerSecond	avgDatarate = client->GetQueueSessionPayloadUp() / client->GetUpStartTimeDelay();
			CMilliSeconds	maxTime = CMilliSeconds(SESSIONMAXTIME) * (avgDatarate / CBytesPerSecond(UPLOAD_CLIENT_DATARATE));
			if (client->GetUpStartTimeDelay() > maxTime)
			{
				if (thePrefs.GetLogUlDlEvents())
					AddDebugLogLine(DLP_DEFAULT, false, _T("%s: Upload session ended due to max time at current datarate. %s (%s)"), client->GetUserName(), CastSecondsToHM(max(maxTime,MIN2MS(5))), CastItoXBytes(avgDatarate));
				timeOver = true;
			}
		}
	}
	// Prepare removal if time is up
	if (timeOver) client->m_dwRemovePending.SetCurrentTick();
	if (uploadSocket)
	{
		if (timeOver)
		{
			uploadSocket->m_eThrottleState = THROTTLE_ENDING; // Ask download throttler to finish off this client
			return (canRemove && uploadSocket->m_bNotNeeded); // Remove if throttler doesn't need the slot anymore
		}
	}
	else return (canRemove && timeOver);
	}
	catch(...)
	{
		AddDebugLogLine(DLP_HIGH, false, _T("CUploadQueue::CheckForTimeOver - An unknown exception occurred"));
		return true; // Force removal due to bug!
	}
	return false; //(timeOver && canRemove);
	// MOD END netfinity
}

/**	
 * \brief	Obtains the number of slots in connecting state
 * \return	uint32 - Number of upload slots in connecting state
 */
uint32
CUploadQueue::MOD_GetConnectingSlotsCount(void) const
{
	uint32	connectingCount = 0;
	POSITION pos = uploadinglist.GetHeadPosition();
	while(pos != NULL)
	{
		CUpDownClient* cur_client = uploadinglist.GetNext(pos);
		if (cur_client->GetUploadState() == US_CONNECTING)
			connectingCount++;
	}
	return connectingCount;
}

/**	
 * \brief	Obtains the current overhead datarate
 * \return	CBytesPerSecond - Overhead in bytes per second
 */
CBytesPerSecond
CUploadQueue::MOD_GetOverheadDatarate(void)
{
	return overheadDatarate;
}